import { world, system } from "@minecraft/server"

world.beforeEvents.worldInitialize.subscribe(i => {
    const blockRegistry = i.blockComponentRegistry
    blockRegistry.registerCustomComponent(`voxel:layer`, {
        onPlace: function placeLayers(e) {
            if (e.block.typeId === "voxel:small_layer") {
                initializeVoxel(e, { x: 4, z: 11 })
                e.block.setPermutation(e.previousBlock)
            }
            else if (e.block.typeId === "voxel:medium_layer") {
                initializeVoxel(e, { x: 2, z: 13 })
                e.block.setPermutation(e.previousBlock)
            }
            else if (e.block.typeId === "voxel:large_layer") {
                initializeVoxel(e, { x: 0, z: 15 })
                e.block.setPermutation(e.previousBlock)
            }
            else if (e.block.typeId === "voxel:chunk") {
                e.block.setPermutation(e.previousBlock)
                system.runTimeout(() => { initializeVoxel(e) }, 2)
            }
        }
    })
})

system.runInterval(() => {
    for (const player of world.getAllPlayers()) {        
        if (!player) return
        if (player.getComponent("inventory").container.getItem(player.selectedSlotIndex)?.typeId === "entity:block_holder_spawn_egg") {
            tracePlayerView(player, false, false, true)
        }
        else tracePlayerView(player)
    }
})

world.afterEvents.entityHitEntity.subscribe(e => {
    const player = e.damagingEntity
    if (player.typeId === "minecraft:player") {
        tracePlayerView(player, true)
    }
})

world.afterEvents.itemUse.subscribe(e => handleVoxelDeletion(e.itemStack, e.source))
world.afterEvents.playerInteractWithEntity.subscribe(e => handleVoxelDeletion(e.itemStack, e.player))

function handleVoxelDeletion(item, player) {
    if (item.typeId === "voxel:delete_all_voxels") {
        player.runCommand(`event entity @e[type=entity:block_holder] despawn`)
        player.onScreenDisplay.setActionBar("§cAll voxels in the world have been removed")
    } 
    else if (item.typeId === "voxel:delete_voxel") tracePlayerView(player, false, false, false, true)
    else tracePlayerView(player, false, true)
}

function tracePlayerView(player, rBlock, placedBlock, takeBlock, deleteAll) {
    const headLocation = player.getHeadLocation()
    const rotation = player.getRotation()

    const pitch = rotation.x * (Math.PI / 180)
    const yaw = rotation.y * (Math.PI / 180)

    const direction = { x: -Math.sin(yaw) * Math.cos(pitch), y: -Math.sin(pitch), z: Math.cos(yaw) * Math.cos(pitch) }
    const stepSize = 1 / 48
    const maxDistance = 2

    if (takeBlock) player.getComponent("inventory").container.setItem(player.selectedSlotIndex)
    for (let step = 0; step <= maxDistance / stepSize; step++) {
        const checkLocation = {
            x: headLocation.x + direction.x * step * stepSize,
            y: (headLocation.y + 0.1) + direction.y * step * stepSize,
            z: headLocation.z + direction.z * step * stepSize
        }
        const entities = player.dimension.getEntities({ type: "entity:block_holder", location: checkLocation, closest: 1, maxDistance: 0.055 })
        if (entities.length > 0) {
            const entityLocation = entities[0].location
            const dx = entityLocation.x - checkLocation.x
            const dy = entityLocation.y - checkLocation.y
            const dz = entityLocation.z - checkLocation.z
            let posOffset = { x: 0, y: 0, z: 0, p: 0 }

            if (Math.abs(dx) > Math.abs(dy) && Math.abs(dx) > Math.abs(dz)) { posOffset.x = dx > 0 ? -1 : 1; posOffset.p = dx > 0 ? 4 : 3 }
            else if (Math.abs(dy) > Math.abs(dx) && Math.abs(dy) > Math.abs(dz)) { posOffset.y = dy > 0 ? -1 : 1; posOffset.p = dy > 0 ? 6 : 5 }
            else { posOffset.z = dz > 0 ? -1 : 1; posOffset.p = dz > 0 ? 2 : 1 }

            if (deleteAll) {
                player.dimension.getEntitiesAtBlockLocation(entityLocation).forEach(e => { if (e.typeId === "entity:block_holder") e.remove() })
                player.onScreenDisplay.setActionBar("§cAll voxels in the block were removed")
                break
            }
            else if (rBlock) {
                entities[0].remove()
            }
            else if (placedBlock) {
                let typeId = player.getComponent("inventory").container.getItem(player.selectedSlotIndex)?.typeId
                if (!typeId || typeId.includes("voxel:")) return
                typeId = (placeableItems.find(f => f.item === typeId)?.block ?? typeId) + " "

                let xyzAxisVals = { a: 1, b: 3, c: 2, d: 0 }
                const pos = { x: entityLocation.x + (posOffset.x / 16), y: entityLocation.y + (posOffset.y / 16), z: entityLocation.z + (posOffset.z / 16) }
                if (player.dimension.getEntities({ type: "entity:block_holder", location: pos, closest: 1, maxDistance: 0.055 })[0]) break

                const block = player.dimension.spawnEntity("entity:block_holder", pos)
                let match;
                for (const item of uniqueBlocks) {
                    match = typeId.includes(item.block)
                    if (match) {
                        if (!item.immutable) {
                            block.setProperty("block:state", item.typeVal ?? 1)
                        }
                        if (item.material) {                            
                            block.setProperty("block:cube", true)
                        }
                        if (item.rotations && !item.notRotable) {
                            xyzAxisVals = item.rotations
                            break
                        }
                        else if (item.notRotable) {
                            xyzAxisVals = undefined
                            break
                        }
                        else break
                    }
                }
                if (!match && solidsNotRotable.some(f => typeId.includes(f))) xyzAxisVals = undefined

                if (xyzAxisVals) {
                    let rotation = xyzAxisVals.a // Norte
                    const playerRotation = player.getRotation().y = (player.getRotation().y + 180) % 360
                    if (playerRotation >= 135 && playerRotation < 225) rotation = xyzAxisVals.b // Sur
                    else if (playerRotation >= 45 && playerRotation < 135) rotation = xyzAxisVals.c // Este
                    else if (playerRotation >= 225 && playerRotation < 315) rotation = xyzAxisVals.d // Oeste
                    block.setProperty("block:rotation", rotation)
                }
                try {
                    block.runCommand(`/replaceitem entity @s slot.weapon.mainhand 0 ${typeId}`)
                    block.runCommand(`/replaceitem entity @s slot.inventory 0 ${typeId}`)
                } catch (e) { }
            }
            else if (takeBlock) {
                const eItem = entities[0].getComponent("inventory").container.getItem(0)
                player.getComponent("inventory").container.setItem(player.selectedSlotIndex, eItem)
            }
            else {
                const hasItem = player.getComponent("inventory").container.getItem(player.selectedSlotIndex)
                player.dimension.spawnEntity("entity:select", { x: entities[0].location.x, y: entities[0].location.y, z: entities[0].location.z }).getComponent("mark_variant").value = hasItem ? posOffset.p : 0
            }
            break
        }
    }
}

function initializeVoxel(i, layerSize) {
    let { x, y, z } = i.block
    let topBlock = i.block.above().isLiquid
    const dimension = i.block.dimension
    y -= 2

    if (!layerSize) {
        const chunkX = Math.floor(x / 16)
        const chunkZ = Math.floor(z / 16)

        const chunkStartX = chunkX * 16
        const chunkStartZ = chunkZ * 16
        const chunkEndX = chunkStartX + 16
        const chunkEndZ = chunkStartZ + 16

        for (let bx = chunkStartX; bx < chunkEndX; bx++) {
            for (let bz = chunkStartZ; bz < chunkEndZ; bz++) {
                for (let by = y; by < y + 16; by++) {
                    if ((bx === chunkStartX && bz === chunkStartZ) || (bx === chunkStartX && bz === chunkEndZ - 1) ||
                        (bx === chunkEndX - 1 && bz === chunkStartZ) || (bx === chunkEndX - 1 && bz === chunkEndZ - 1) ||
                        ((by === y + 1 || by === y + 15) && (bx === chunkStartX || bx === chunkEndX - 1 || bz === chunkStartZ || bz === chunkEndZ - 1))) {
                        dimension.spawnParticle("select:border", { x: bx + 0.5, y: by, z: bz + 0.5 })
                    }
                    const block = dimension.getBlock({ x: bx, y: by, z: bz })
                    let typeId = block.typeId
                    typeId = (replicaBlocks.find(f => f.block === typeId)?.replica ?? typeId) + " "
                    if (block.isAir || (topBlock && block.isLiquid) || typeId.includes("voxel:")) continue

                    let cardinalDir = { a: 3, b: 1, c: 2, d: 0 }
                    let facingDir = { a: 1, b: 3, c: 0, d: 2 }
                    let weirdo_dir = { a: 3, b: 1, c: 0, d: 2 }
                    let direction = { a: 3, b: 0, c: 3, d: 2 }
                    let rotation;
                    switch (block.permutation.getState("cardinal_direction")) {
                        case 2: rotation = cardinalDir.a; break; case 3: rotation = cardinalDir.b; break; case 4: rotation = cardinalDir.c; break; case 5: rotation = cardinalDir.d; break
                    }
                    switch (block.permutation.getState("facing_direction")) {
                        case 2: rotation = facingDir.a; break; case 3: rotation = facingDir.b; break; case 4: rotation = facingDir.c; break; case 5: rotation = facingDir.d; break
                    }
                    switch (block.permutation.getState("weirdo_direction")) {
                        case 0: rotation = weirdo_dir.a; break; case 1: rotation = weirdo_dir.b; break; case 2: rotation = weirdo_dir.c; break; case 3: rotation = weirdo_dir.d; break
                    }
                    switch (block.permutation.getState("direction")) {
                        case 0: rotation = direction.a; break; case 1: rotation = direction.b; break; case 2: rotation = direction.c; break; case 3: rotation = direction.d; break
                    }
                    let isSurrounded = true
                    const positions = [{ x: bx + 1, y: by, z: bz }, { x: bx - 1, y: by, z: bz }, { x: bx, y: by + 1, z: bz }, { x: bx, y: by - 1, z: bz }, { x: bx, y: by, z: bz + 1 }, { x: bx, y: by, z: bz - 1 }]
                    for (let pos of positions) {
                        if (pos.x < chunkStartX || pos.x >= chunkEndX || pos.z < chunkStartZ || pos.z >= chunkEndZ || pos.y < y || pos.y >= y + 16) {
                            isSurrounded = false; break
                        }
                        const neighbor = dimension.getBlock(pos)
                        if (!neighbor || (topBlock && neighbor.isLiquid) || uniqueBlocks.some(f => (neighbor.typeId + " ").includes(f.block))) {
                            isSurrounded = false; break
                        }
                    }
                    if (isSurrounded) continue

                    const miniblock = dimension.spawnEntity("entity:block_holder", { x: (x + 0.03072) + (bx - chunkStartX) / 16, y: (y + 2.031) + (by - y) / 16, z: (z + 0.03072) + (bz - chunkStartZ) / 16 })
                    for (const item of uniqueBlocks) {
                        const match = typeId.includes(item.block)
                        if (match) {
                            if (!item.immutable) {
                                miniblock.setProperty("block:state", item.typeVal ?? 1)
                            }
                            if (item.material) {
                                miniblock.setProperty("block:cube", true)
                                break
                            }
                            else break
                        }
                    }
                    if (rotation) miniblock.setProperty("block:rotation", rotation)
                    try {
                        miniblock.runCommand(`/replaceitem entity @s slot.weapon.mainhand 0 ${typeId}`)
                        miniblock.runCommand(`/replaceitem entity @s slot.inventory 0 ${typeId}`)
                    } catch (e) { }
                }
            }
        }
    }
    else for (let i = layerSize.x; i <= layerSize.z; i++) {
        for (let j = layerSize.x; j <= layerSize.z; j++) {
            const typeId = "minecraft:grass_block "
            const miniblock = dimension.spawnEntity("entity:block_holder", { x: (x + 0.03072) + i / 16, y: y + 2.031, z: (z + 0.03072) + j / 16 })
            /*for (const item of uniqueBlocks) {
                if (typeId.includes(item.block) && !item.immutable) { miniblock.setProperty("block:state", item.typeVal ?? 1); break }
            }*/
            try {
                miniblock.runCommand(`/replaceitem entity @s slot.weapon.mainhand 0 ${typeId}`)
                miniblock.runCommand(`/replaceitem entity @s slot.inventory 0 ${typeId}`)
            } catch (e) { }
        }
    }
}

const replicaBlocks = [
    { block: "minecraft:water", replica: "replica:water" },
    { block: "minecraft:lava", replica: "replica:lava" },
    { block: "minecraft:flowing_water", replica: "replica:water" },
    { block: "minecraft:flowing_lava", replica: "replica:lava" },
    { block: "minecraft:portal", replica: "replica:nether_portal" },
    { block: "minecraft:end_portal", replica: "replica:end_portal" },
]

const placeableItems = [
    { item: "minecraft:water_bucket", block: "replica:water" },
    { item: "minecraft:lava_bucket", block: "replica:lava" },
]

const solidsNotRotable = [
    "stone ",
    "granite",
    "diorite",
    "andesite",
    "grass_block",
    "scaffolding",
    "dirt",
    "podzol",
    "_planks",
    "bedrock",
    "sand ",
    "gravel",
    "_ore ",
    "_log",
    "stripped_",
    "_wood ",
    "_leaves",
    "sponge",
    "dispenser",
    "noteblock",
    "_wool",
    "_quartz",
    "quartz_",
    "tnt",
    "bookshelf",
    "obsidian",
    ":ice ",
    "_ice",
    "clay",
    "bricks",
    "_block",
    "concrete",
]

const uniqueBlocks = [
    { block: ":air " },
    { block: "voxel:" },
    { block: "vine" },
    { block: "fern" },
    { block: "chest", immutable: true, rotations: { a: 3, b: 1, c: 0, d: 2 } },
    { block: "_stairs", immutable: true, rotations: { a: 2, b: 0, c: 3, d: 1 } },
    { block: "seagrass" },
    { block: "_wall", typeVal: 5 },
    { block: "ladder" },
    { block: "_sapling" },

    { block: "short_grass" },
    { block: "banner", typeVal: 9 },
    { block: "sweet_berry_bush" },
    { block: "deadbush" },
    { block: "torch" },
    { block: "_torch" },
    { block: ":glass ", immutable: true, notRotable: true },
    { block: "deadbush" },
    { block: "_propagule", material: true },
    { block: "waterlily" },
    { block: "tall_grass" },
    { block: "large_fern" },
    { block: "campfire" },
    { block: "bamboo", typeVal: 2 },
    { block: "bell", material: true },

    { block: "flower_pot" },
    { block: "peony" },
    { block: "dandelion" },
    { block: "poppy" },
    { block: "torchflower" },
    { block: "rose_bush" },
    { block: "sunflower" },
    { block: "lilac" },
    { block: "blue_orchid" },
    { block: "allium" },
    { block: "azure_bluet" },
    { block: "oxeye_daisy" },
    { block: "cornflower" },
    { block: "lily_of_the_valley" },
    { block: "wither_rose" },
    { block: "_tulip" },
    { block: "pink_petals" },
    { block: "pointed_dripstone" },
    { block: "glow_lichen" },
    { block: "_head", immutable: true },
    { block: "_skull", immutable: true },

    { block: ":chain " },
    { block: "iron_bars", typeVal: 2, rotations: { a: 2, b: 2, c: 1, d: 1 } },
    { block: "fence", immutable: true, rotations: { a: 2, b: 2, c: 1, d: 1 } },
    { block: "anvil", immutable: true, rotations: { a: 2, b: 2, c: 1, d: 1 } },

    { block: "moss_carpet", typeVal: 10, notRotable: true },
    { block: "_slab", typeVal: 8, notRotable: true, material: true },
    { block: "_pressure_plate", typeVal: 10, notRotable: true },
    { block: "_carpet", typeVal: 6, notRotable: true },
    { block: "snow_layer", typeVal: 7, notRotable: true },
    { block: "lantern" },
    { block: "_sign", material: true },
    { block: "soul_lantern" },
    { block: "brewing_stand" },
    { block: "carrots" },
    { block: "rail " },
    { block: "potatoes" },
    { block: "wheat" },
    { block: "seeds" },
    { block: "beetroots" },
    { block: "melon_stem" },
    { block: "pumpkin_stem" },
    { block: "candle" },
    { block: "lever" },
    { block: "armor_stand", typeVal: 3 },
    { block: "nether_wart" },
    { block: "coral " },
    { block: "coral_fan " },
    { block: ":comparator " },
    { block: ":repeater " },
    { block: ":redstone " },
    { block: ":bed " },
    { block: ":web " },

    { block: "stonecutter", immutable: true },
    { block: "_door", typeVal: 3, rotations: { a: 2, b: 2, c: 1, d: 1 } },
    { block: "trapdoor", typeVal: 10 },
    { block: "_button", typeVal: 11 },

    { block: "_stained_glass ", immutable: true, notRotable: true },
    { block: "glass_pane ", typeVal: 4, material: true, rotations: { a: 2, b: 2, c: 1, d: 1 } },

    { block: "structure_void", immutable: true },
    { block: "light_block" },
    { block: "barrier" }
]